/*
 * Tabs are 4 symbols wide due to stupid Windows function names...
 * ---------------------------------------------------------------
 *
 * Copyright (c) 2010 KSI Labs, LLC. All rights reserved.
 * Written by Sergey Kubushyn <ksi@koi8.net>, <ksi@ksilabs.com>.
 *
 * File: centipede.h
 *
 * This is main include file for WDM driver for KSI Labs, LLC CENTIPEDE
 * PCI Board. Most of the data structures, flags etc. are declared here.
 *
 * Please read documentation or go to http://www.ksilabs.com for further info.
 */

/*
 * 						LICENSE and Terms of Use.
 *
 * First of all, this is NOT GPL or Free Software. It is open so you have
 * all the source code for the Engine. There is also VHDL source code for
 * CENTIPEDE board CPLD that is available from KSI Labs, LLC website,
 * http://www.ksilabs.com.
 *
 * As for this particular code, Windows Mach3 Engine, you are free to use it,
 * make any modifications to customize it for your own needs, fix inevitable
 * bugs, port it to other CNC software and do anything else as long as your
 * derivative work that is based on this code or uses parts or pieces thereof
 * is made to be used with any KSI Labs, LLC products.
 *
 * We do NOT grant you a permission to use any parts of this software in any
 * products that are designed for products not manufactured or distributed
 * by KSI Labs, LLC. We might give such a permission on case-by-case basis
 * but it is not legal to use this software or bits and pieces thereof with
 * products not coming from KSI Labs, LLC without our explicit written
 * permission.
 *
 * The original copyright notices must be included in all derivative works.
 * You do NOT have to ask our permission to make any modifications to this
 * code and you are not under obligation to give us any of your work back
 * as long as you are using it for your own needs.
 *
 * However we do NOT give our permission to distribute your derivative work
 * unless all original copyrights are retained, all the conditions of this
 * license are met (use only with KSI Labs, LLC products) and full source
 * code for such derivative work is available for free to everybody without
 * any further conditions including but not limited to signing any kinds of
 * NDA. We explicitely deny permission to distribute any derivative work
 * based on this code with closed source binary-only fragments or include
 * bits or pieces of this code in such products.
 *
 * CNC software like Mach3 and likes is excluded from these requirements as
 * long as it does not include parts or pieces of this software into the
 * close source proprietary application proper. It can use shared memory
 * buffer, IOCTL calls, etc. to communicate with this Engine or the one based
 * on it and keep all the software making those calls or working on that
 * buffer close source and proprietary as long as those calls/structures are
 * documented and that documentation is freely available to everybody without
 * any further conditions including but no limited to signing any kind of NDA.
 */

#if !defined(_CPDDRV_H_)
#define _CPDDRV_H_

//
// In this driver we are using a safe version vsnprintf, which is
// RtlStringCbVPrintfA. To use strsafe function on Win2K, we have to
// define NTSTRSAFE_LIB before including this header file and explicitly
// link to ntstrsafe.lib. If the driver is just target for XP and above,
// there is no define NTSTRSAFE_LIB and link to the lib.
//
//#define NTSTRSAFE_LIB
//#include <ntstrsafe.h>

//
// InterlockedOr is not defined in Win2K header files.
// If not defined, we will define it to use intrinsic function.
//
#if !defined(InterlockedOr) && (_WIN32_WINNT==0x0500)
#define InterlockedOr _InterlockedOr
#endif

#define CPDDRV_POOL_TAG (ULONG) 'DDPC'
#define CPDDRV_FDO_INSTANCE_SIGNATURE (ULONG) 'odFT'

#define DRIVERVERSION			580

//
// Bit Flag Macros
//

// CPD Board flags
#define CPD_IRQ_IN_USE	0x00000008

//
// These are the states FDO transition to upon
// receiving a specific PnP Irp. Refer to the PnP Device States
// diagram in DDK documentation for better understanding.
//

typedef enum _DEVICE_PNP_STATE {
	NotStarted = 0,         // Not started yet
	Started,                // Device has received the START_DEVICE IRP
	StopPending,            // Device has received the QUERY_STOP IRP
	Stopped,                // Device has received the STOP_DEVICE IRP
	RemovePending,          // Device has received the QUERY_REMOVE IRP
	SurpriseRemovePending,  // Device has received the SURPRISE_REMOVE IRP
	Deleted                 // Device has received the REMOVE_DEVICE IRP
} DEVICE_PNP_STATE;

//
// This enum used in implementing the wait-lock
//
typedef enum _WAIT_REASON {
	DEV_REMOVE = 0,
	DEV_STOP
} WAIT_REASON;

#define INITIALIZE_PNP_STATE(_Data_)    \
		(_Data_)->DevicePnPState =  NotStarted;\
		(_Data_)->PreviousPnPState = NotStarted;

#define SET_NEW_PNP_STATE(_Data_, _state_) \
		(_Data_)->PreviousPnPState =  (_Data_)->DevicePnPState;\
		(_Data_)->DevicePnPState = (_state_);

#define RESTORE_PREVIOUS_PNP_STATE(_Data_)   \
		(_Data_)->DevicePnPState =   (_Data_)->PreviousPnPState;\

typedef enum _QUEUE_STATE {
	HoldRequests = 0,
	AllowRequests,
	FailRequests
} QUEUE_STATE;

//
// Driver FSM states
//
typedef enum _DRV_STATE {
	DrvNotReady = 0,
	DrvEStopped,
	DrvInProgram,
	DrvJogging,
	DrvProbing,
	DrvHomingPrePass,
	DrvWaitingForHome,
	DrvHoming1stPass,
	DrvWaitingForDeref,
	DrvHomingDeref,
	DrvJogFinishing,
	DrvProgramFinishing,
	DrvDwelling,
	DrvIdle
} DRV_STATE;

typedef enum _AXIS_TYPE {
	Single = 0,
	Master,
	Slave
} AXIS_TYPE;

typedef struct _AXIS_DATA {
	AXIS_TYPE	Type;
	INT			Slave;
	INT			Master;
	INT32		HitIDX;
	INT32		PendingIDX;
	INT			QuantCntr;
	INT32		TripPoint;
	UINT32		HomePreMask;
	BOOLEAN		Active;
	BOOLEAN		Hit;
} AXIS_DATA, *PAXIS_DATA;

typedef struct _MPG_DATA {
	BOOLEAN		StateA;
	BOOLEAN		StateB;
} MPG_DATA, *PMPG_DATA;

//
// General purpose workitem context used in dispatching work to
// system worker thread to be executed at PASSIVE_LEVEL.
//
typedef struct _WORKER_ITEM_CONTEXT {
	PIO_WORKITEM   WorkItem;
	PVOID          Callback; // Callback pointer
	PVOID          Argument1;
	PVOID          Argument2;
} WORKER_ITEM_CONTEXT, *PWORK_ITEM_CONTEXT;

typedef struct TimeData		MACH_TD_STRUCT;
typedef struct TimeData 	*PMACH_TD_STRUCT;

//
// The device extension for the device object
//
typedef struct _FDO_DATA
{
	ULONG                   Signature;
	PDEVICE_OBJECT          Self;
	PDEVICE_OBJECT          UnderlyingPDO;
	PDEVICE_OBJECT          NextLowerDriver;
	DEVICE_PNP_STATE        DevicePnPState;
	DEVICE_PNP_STATE        PreviousPnPState;
	UNICODE_STRING          InterfaceName;
	QUEUE_STATE             QueueState;
	LIST_ENTRY              NewRequestsQueue;
	KSPIN_LOCK              QueueLock;
	KEVENT                  RemoveEvent;
	KEVENT                  StopEvent;
	ULONG                   OutstandingIO;

	//////////////////////
	// Power Management //
	//////////////////////

	SYSTEM_POWER_STATE      SystemPowerState;
	DEVICE_POWER_STATE      DevicePowerState;
	PIRP                    PendingSIrp;

	//////////////////////////////
	// Hardware-specific fields //
	//////////////////////////////

	// Pointer to Mach3 Shared Memory Block where all the queues etc. live
	// and to TimeData struct that is mapped on SOFT_SAMPLE_BUFF IOCTL.
	volatile PMACH_STRUCT		MachSharedBlock;
	volatile PMACH_TD_STRUCT	MachTDBlock;

	// IRPs that brought IOCTL to map MachSharedBuffer and TimeData struct
	PIRP					MapBufIrp;
	PIRP					MapTDIrp;

	///////////////////
	// Configuration //
	///////////////////

	// HW Resources
	ULONG					Flags;
	PHYSICAL_ADDRESS        MemPhysAddress;
	PKINTERRUPT             Interrupt;
	UCHAR                   InterruptLevel;
	ULONG                   InterruptVector;
	KAFFINITY               InterruptAffinity;
	ULONG                   InterruptMode;
	PARALLEL_PORT_INFORMATION	PPortInfo[2];
	BOOLEAN					Port1isOK;
	BOOLEAN					Port1isInput;
	BOOLEAN					Port1On;
	BOOLEAN					Port2isOK;
	BOOLEAN					Port2isInput;
	BOOLEAN					Port2On;
	BOOLEAN					PortsMapOK;
	UINT32					PortsInBuf;
	UINT32					CPDInBuf;
	UINT32					PortsOutBuf;
	UINT32					CPDOutBuf;
	volatile PCPD_REGS		CPDRegsBase;
	UINT32					IRQMask;
	INT						TmrFreq;

	// Driver state
	BOOLEAN					Running;
	DRV_STATE				DrvState;
	BOOLEAN					StateChange;

	// Internal driver data structures
	INT						JogUnitsPerStep;
	INT						CyclesPerStat;
	INT						StatCntr;
	AXIS_DATA				AxisData[6];
	MPG_DATA				MPGData[3];

	// I/O call tables
	pGetSigFunc				GetSigMap[nSigs];
	pPutSigFunc				PutSigMap[nSigsOut];
	pGetSigFunc				GetMPGSigMap[6];

	// spin locks for protecting misc variables
	KSPIN_LOCK              Lock;

} FDO_DATA, *PFDO_DATA;

//
// Function prototypes
//
NTSTATUS DriverEntry(PDRIVER_OBJECT  DriverObject, PUNICODE_STRING RegistryPath);
NTSTATUS SystemControl(PDEVICE_OBJECT DeviceObject, PIRP Irp);
NTSTATUS AddDevice(PDRIVER_OBJECT DriverObject, PDEVICE_OBJECT PhysicalDeviceObject);
NTSTATUS DispatchPnp(PDEVICE_OBJECT DeviceObject, PIRP Irp);
NTSTATUS DispatchIoctl(PDEVICE_OBJECT DeviceObject, PIRP Irp);
NTSTATUS StartDevice(PFDO_DATA FdoData, PIRP Irp);
NTSTATUS Create(PDEVICE_OBJECT DeviceObject, PIRP Irp);
NTSTATUS Close(PDEVICE_OBJECT DeviceObject, PIRP Irp);
NTSTATUS Cleanup(PDEVICE_OBJECT DeviceObject, PIRP Irp);
VOID Unload(PDRIVER_OBJECT DriverObject);
NTSTATUS ForwardAndForget(PFDO_DATA FdoData, PIRP Irp);
NTSTATUS SendIrpSynchronously(PDEVICE_OBJECT DeviceObject, PIRP Irp);
LONG IoIncrement(PFDO_DATA FdoData);
LONG IoDecrement(PFDO_DATA FdoData);
VOID ReleaseAndWait(PFDO_DATA FdoData, ULONG ReleaseCount, WAIT_REASON Reason);
VOID ProcessQueuedRequests(PFDO_DATA FdoData);
NTSTATUS ReturnResources(PDEVICE_OBJECT DeviceObject);
NTSTATUS DispatchPnpComplete(PDEVICE_OBJECT DeviceObject, PIRP Irp, PVOID Context);
NTSTATUS QueueRequest(PFDO_DATA FdoData, PIRP Irp);
VOID CancelQueued(PDEVICE_OBJECT DeviceObject, PIRP Irp);
VOID CancelRoutineForMapBufIoctlIrp(PDEVICE_OBJECT DeviceObject, PIRP Irp);
VOID CancelRoutineForMapTDIoctlIrp(PDEVICE_OBJECT DeviceObject, PIRP Irp);
NTSTATUS QueuePassiveLevelCallback(PFDO_DATA FdoData,
		PIO_WORKITEM_ROUTINE CallbackFunction, PVOID Context1, PVOID Context2);


// utils.c
BOOLEAN ReadRegistryValue(PFDO_DATA FdoData, PWCHAR Name, PLONG Value);
BOOLEAN WriteRegistryValue(PFDO_DATA FdoData, PWCHAR Name, ULONG Value);
PCHAR PnPMinorFunctionString(UCHAR MinorFunction);
NTSTATUS GetPPortInfo(PFDO_DATA FdoData, UINT PortNum);

#if defined(WIN2K)
NTKERNELAPI VOID ExFreePoolWithTag(PVOID P, ULONG Tag);

//
// This value should be returned from completion routines to continue
// completing the IRP upwards. Otherwise, STATUS_MORE_PROCESSING_REQUIRED
// should be returned.
//
#define STATUS_CONTINUE_COMPLETION      STATUS_SUCCESS
#endif

// IOCTLs (Almost all ot which are not implemented :))
#define MACH1_PREPTIMER			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x800, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define MACH1_STOPTIMER			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x801, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define MACH1_GETCOUNT			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x802, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define MACH1_GETMAXTIME		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x803, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define MACH1_GETMINTIME		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x804, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define MACH1_STARTTIMER		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x805, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_GET_STATUS			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x806, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_CONTROL_OUT		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x807, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_SET_FREQUENCY		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x808, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_ACTIVATE			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x809, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_GET_TIMES			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x80a, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_GET_DATA			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x80b, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_ADD_TO_BUFFER		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x80c, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_GET_LOADINDEX		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x80d, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_GET_CONTROLINDEX	CTL_CODE(FILE_DEVICE_UNKNOWN, 0x80e, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_SET_PORT			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x80f, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_LOAD_MASKS			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x810, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_GET_ACTIVE			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x811, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_GETINDEXS			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x812, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_SETREVERSALS		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x813, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_CONTROL_IN			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x814, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_CONTROL_JOGON		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x815, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_CONTROL_JOGOFF		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x816, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_SETNEGS			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x817, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_GETIDLE			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x818, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_EMERGENCY			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x819, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_OVERRIDE			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x81a, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_GOHOME				CTL_CODE(FILE_DEVICE_UNKNOWN, 0x81b, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_PAUSE				CTL_CODE(FILE_DEVICE_UNKNOWN, 0x81c, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_HALT				CTL_CODE(FILE_DEVICE_UNKNOWN, 0x81d, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_LASER				CTL_CODE(FILE_DEVICE_UNKNOWN, 0x81e, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_SPACE				CTL_CODE(FILE_DEVICE_UNKNOWN, 0x81f, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_GATE				CTL_CODE(FILE_DEVICE_UNKNOWN, 0x820, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define SOFT_ALLOCATE_BUFF		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x821, METHOD_OUT_DIRECT,FILE_ANY_ACCESS)
#define SOFT_SAMPLE_BUFF		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x822, METHOD_OUT_DIRECT,FILE_ANY_ACCESS)
#define SOFT_SAMPLE_ON			CTL_CODE(FILE_DEVICE_UNKNOWN, 0x823, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define MACH1_RELEASEBUFFERS	CTL_CODE(FILE_DEVICE_UNKNOWN, 0x824, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define MACH1_DRIVERVERSION		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x825, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define MACH1_WAITFORTIMER		CTL_CODE(FILE_DEVICE_UNKNOWN, 0x826, METHOD_BUFFERED, FILE_ANY_ACCESS)

#endif  // _CPDDRV_H_
