/*
 * Tabs are 4 symbols wide due to stupid Windows function names...
 * ---------------------------------------------------------------
 *
 * Copyright (c) 2010 KSI Labs, LLC. All rights reserved.
 * Written by Sergey Kubushyn <ksi@koi8.net>, <ksi@ksilabs.com>.
 *
 * File: utils.c
 *
 * This is a set of utility functions for CENTIPEDE driver. Registry access,
 * DebugPrint, and likes. See centipede.c file for more detailed descriptio
 * of CENTIPEDE board and driver.
 */

/*
 * 						LICENSE and Terms of Use.
 *
 * First of all, this is NOT GPL or Free Software. It is open so you have
 * all the source code for the Engine. There is also VHDL source code for
 * CENTIPEDE board CPLD that is available from KSI Labs, LLC website,
 * http://www.ksilabs.com.
 *
 * As for this particular code, Windows Mach3 Engine, you are free to use it,
 * make any modifications to customize it for your own needs, fix inevitable
 * bugs, port it to other CNC software and do anything else as long as your
 * derivative work that is based on this code or uses parts or pieces thereof
 * is made to be used with any KSI Labs, LLC products.
 *
 * We do NOT grant you a permission to use any parts of this software in any
 * products that are designed for products not manufactured or distributed
 * by KSI Labs, LLC. We might give such a permission on case-by-case basis
 * but it is not legal to use this software or bits and pieces thereof with
 * products not coming from KSI Labs, LLC without our explicit written
 * permission.
 *
 * The original copyright notices must be included in all derivative works.
 * You do NOT have to ask our permission to make any modifications to this
 * code and you are not under obligation to give us any of your work back
 * as long as you are using it for your own needs.
 *
 * However we do NOT give our permission to distribute your derivative work
 * unless all original copyrights are retained, all the conditions of this
 * license are met (use only with KSI Labs, LLC products) and full source
 * code for such derivative work is available for free to everybody without
 * any further conditions including but not limited to signing any kinds of
 * NDA. We explicitely deny permission to distribute any derivative work
 * based on this code with closed source binary-only fragments or include
 * bits or pieces of this code in such products.
 *
 * CNC software like Mach3 and likes is excluded from these requirements as
 * long as it does not include parts or pieces of this software into the
 * close source proprietary application proper. It can use shared memory
 * buffer, IOCTL calls, etc. to communicate with this Engine or the one based
 * on it and keep all the software making those calls or working on that
 * buffer close source and proprietary as long as those calls/structures are
 * documented and that documentation is freely available to everybody without
 * any further conditions including but no limited to signing any kind of NDA.
 */

#include "precomp.h"

#define PARPORT_NAME			L"\\Device\\ParallelPort"
#define STR_ZERO				L"0"
#define STR_ONE					L"1"
#define STR_TWO					L"2"
#define STR_THREE				L"3"

// Uncomment this if you want debugging info
//
//#define DEBUG	1

#ifdef DEBUG
ULONG DebugLevel = LOUD;
ULONG DebugFlag = 0xFFFFFFFF;
#endif

#ifdef ALLOC_PRAGMA
#pragma alloc_text (PAGE, ReadRegistryValue)
#pragma alloc_text (PAGE, WriteRegistryValue)
#endif

// ***************************************************************************
// This one gets PARALLEL_PORT_INFORMATION structs for a parallel port       *
// in the system if it exists. PortNum is ZERO-based. Returns STATUS_SUCCESS *
// if port exists and info was retrived, failure code otherwise.             *
// ***************************************************************************
NTSTATUS GetPPortInfo(PFDO_DATA FdoData, UINT PortNum)
{
	NTSTATUS					status = STATUS_SUCCESS;
	UNICODE_STRING 				PPDevName;
	PFILE_OBJECT				pFileObj = NULL;
	PDEVICE_OBJECT				pDevObj = NULL;
	PIRP						pIrp = NULL;
	KEVENT						WaitEvent;
	IO_STATUS_BLOCK				IrpCreationStatus;

	switch (PortNum) {
		case 0:
			RtlInitUnicodeString(&PPDevName, PARPORT_NAME STR_ZERO);
			break;
		case 1:
			RtlInitUnicodeString(&PPDevName, PARPORT_NAME STR_ONE);
			break;
/*
		case 2:
			RtlInitUnicodeString(&PPDevName, PARPORT_NAME STR_TWO);
			break;

		case 3:
			RtlInitUnicodeString(&PPDevName, PARPORT_NAME STR_THREE);
			break;
*/
		default:
			return STATUS_INVALID_PARAMETER;
	}

	//
	// Try to get a pointer to Parallel Port device
	//
	status = IoGetDeviceObjectPointer(&PPDevName, FILE_ALL_ACCESS,
							&pFileObj, &pDevObj);

	if (status != STATUS_SUCCESS) return status;

	//
	// Now try to get port info for this port with internal IOCTL.
	//
	KeInitializeEvent(&WaitEvent, NotificationEvent, FALSE);

	pIrp = IoBuildDeviceIoControlRequest (
				IOCTL_INTERNAL_GET_PARALLEL_PORT_INFO,
				pDevObj,
				&FdoData->PPortInfo[PortNum],
				sizeof(PARALLEL_PORT_INFORMATION),
				&FdoData->PPortInfo[PortNum],
				sizeof(PARALLEL_PORT_INFORMATION),
				TRUE,
				&WaitEvent,
				&IrpCreationStatus
			);

	if (pIrp == NULL) {
		DebugPrint(ERROR, DBG_INIT, "GetPPortInfo(): Allocation of IRP (PORT_GET_INFO) failed\n");
		ObDereferenceObject(pFileObj);
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	status = IoCallDriver(pDevObj, pIrp);
	if (status == STATUS_PENDING) {
		KeWaitForSingleObject(&WaitEvent, Executive, KernelMode, FALSE, NULL);
	}

	status = IrpCreationStatus.Status;
	if (status != STATUS_SUCCESS) {
		DebugPrint(ERROR, DBG_INIT, "GetPPortInfo(): pIrpCreationStatus -- Failed\n");
		ObDereferenceObject(pFileObj);
		return status;
	}

	ObDereferenceObject(pFileObj);

	return status;
}


// *****************************************************
// * Can be used to read any REG_DWORD registry value  *
// * stored under Device Parameter. Not used as of now *
// * but might be used later for something.            *
// *****************************************************
BOOLEAN ReadRegistryValue(PFDO_DATA FdoData, PWCHAR Name, PLONG Value)
{
	HANDLE			hKey = NULL;
	NTSTATUS		status;
	BOOLEAN			retValue = FALSE;
	ULONG			length;
	PKEY_VALUE_FULL_INFORMATION	fullInfo;
	UNICODE_STRING  valueName;

	PAGED_CODE();

	DebugPrint(TRACE, DBG_INIT, "-->ReadRegistryValue \n");

	*Value = 0;

	status = IoOpenDeviceRegistryKey(FdoData->UnderlyingPDO,
						PLUGPLAY_REGKEY_DEVICE, STANDARD_RIGHTS_ALL, &hKey);

	if (NT_SUCCESS(status)) {

		RtlInitUnicodeString(&valueName, Name);

		length = sizeof(KEY_VALUE_FULL_INFORMATION)
							+ valueName.MaximumLength
							+ sizeof(ULONG);

		fullInfo = ExAllocatePoolWithTag(PagedPool, length, CPDDRV_POOL_TAG);

		if (fullInfo) {

			status = ZwQueryValueKey(hKey, &valueName, KeyValueFullInformation,
										fullInfo, length, &length);

			if (NT_SUCCESS(status)) {

				ASSERT(sizeof(ULONG) == fullInfo->DataLength);
				RtlCopyMemory (Value,
								((PUCHAR)fullInfo) + fullInfo->DataOffset,
								fullInfo->DataLength);
				retValue = TRUE;
			}

			ExFreePool(fullInfo);
		}

		ZwClose(hKey);
	}

	DebugPrint(TRACE, DBG_INIT, "<--ReadRegistryValue %ws %d \n",
									Name, *Value);
	return retValue;
}


// ************************************************************
// * Can be used to write any REG_DWORD registry value stored *
// * under Device Parameter. Returns TRUE on success.         *
// ************************************************************
BOOLEAN WriteRegistryValue(PFDO_DATA FdoData, PWCHAR Name, ULONG Value)
{
	HANDLE          hKey = NULL;
	NTSTATUS        status;
	UNICODE_STRING  valName;
	BOOLEAN         retValue = FALSE;

	PAGED_CODE();

	DebugPrint(TRACE, DBG_INIT, "Entered WriteRegistryValue\n");

	status = IoOpenDeviceRegistryKey(FdoData->UnderlyingPDO,
						PLUGPLAY_REGKEY_DEVICE, STANDARD_RIGHTS_ALL, &hKey);

	if (NT_SUCCESS(status)) {

		RtlInitUnicodeString (&valName, Name);

		status = ZwSetValueKey(hKey, &valName, 0, REG_DWORD, &Value, sizeof(Value));

		if (NT_SUCCESS(status))
			retValue = TRUE;

		ZwClose (hKey);
	}

	return retValue;
}


VOID DebugPrint(ULONG PrintLevel, ULONG PrintFlag, PCCHAR Message, ...)
{
#ifdef DEBUG
#define     TEMP_BUFFER_SIZE        1024
	va_list		list;
	UCHAR		debugMessageBuffer[TEMP_BUFFER_SIZE];
	NTSTATUS	status;

	va_start(list, Message);

	if (Message) {
		//
		// Using new safe string functions instead of _vsnprintf. This
		// function takes care of NULL terminating if the message is longer
		// than the buffer.
		//
		status = RtlStringCbVPrintfA(debugMessageBuffer, sizeof(debugMessageBuffer),
									Message, list);
		if (!NT_SUCCESS(status)) {

			DbgPrint("CENTIPEDE: RtlStringCbVPrintfA failed %x\n", status);
			return;
		}

		if (PrintLevel <= INFO || (PrintLevel <= DebugLevel &&
				((PrintFlag & DebugFlag) == PrintFlag))) {

			DbgPrint("CENTIPEDE:%s", debugMessageBuffer);
		}
	}

	va_end(list);

	return;
#endif
}


// **************************************************
// * Debug strings mappings, PnP IRP Minor Function *
// **************************************************
PCHAR PnPMinorFunctionString(UCHAR MinorFunction)
{
#ifdef DEBUG
	switch (MinorFunction)
	{
		case IRP_MN_START_DEVICE:
			return "IRP_MN_START_DEVICE";
		case IRP_MN_QUERY_REMOVE_DEVICE:
			return "IRP_MN_QUERY_REMOVE_DEVICE";
		case IRP_MN_REMOVE_DEVICE:
			return "IRP_MN_REMOVE_DEVICE";
		case IRP_MN_CANCEL_REMOVE_DEVICE:
			return "IRP_MN_CANCEL_REMOVE_DEVICE";
		case IRP_MN_STOP_DEVICE:
			return "IRP_MN_STOP_DEVICE";
		case IRP_MN_QUERY_STOP_DEVICE:
			return "IRP_MN_QUERY_STOP_DEVICE";
		case IRP_MN_CANCEL_STOP_DEVICE:
			return "IRP_MN_CANCEL_STOP_DEVICE";
		case IRP_MN_QUERY_DEVICE_RELATIONS:
			return "IRP_MN_QUERY_DEVICE_RELATIONS";
		case IRP_MN_QUERY_INTERFACE:
			return "IRP_MN_QUERY_INTERFACE";
		case IRP_MN_QUERY_CAPABILITIES:
			return "IRP_MN_QUERY_CAPABILITIES";
		case IRP_MN_QUERY_RESOURCES:
			return "IRP_MN_QUERY_RESOURCES";
		case IRP_MN_QUERY_RESOURCE_REQUIREMENTS:
			return "IRP_MN_QUERY_RESOURCE_REQUIREMENTS";
		case IRP_MN_QUERY_DEVICE_TEXT:
			return "IRP_MN_QUERY_DEVICE_TEXT";
		case IRP_MN_FILTER_RESOURCE_REQUIREMENTS:
			return "IRP_MN_FILTER_RESOURCE_REQUIREMENTS";
		case IRP_MN_READ_CONFIG:
			return "IRP_MN_READ_CONFIG";
		case IRP_MN_WRITE_CONFIG:
			return "IRP_MN_WRITE_CONFIG";
		case IRP_MN_EJECT:
			return "IRP_MN_EJECT";
		case IRP_MN_SET_LOCK:
			return "IRP_MN_SET_LOCK";
		case IRP_MN_QUERY_ID:
			return "IRP_MN_QUERY_ID";
		case IRP_MN_QUERY_PNP_DEVICE_STATE:
			return "IRP_MN_QUERY_PNP_DEVICE_STATE";
		case IRP_MN_QUERY_BUS_INFORMATION:
			return "IRP_MN_QUERY_BUS_INFORMATION";
		case IRP_MN_DEVICE_USAGE_NOTIFICATION:
			return "IRP_MN_DEVICE_USAGE_NOTIFICATION";
		case IRP_MN_SURPRISE_REMOVAL:
			return "IRP_MN_SURPRISE_REMOVAL";
		case IRP_MN_QUERY_LEGACY_BUS_INFORMATION:
			return "IRP_MN_QUERY_LEGACY_BUS_INFORMATION";
		default:
			return "unknown_pnp_irp";
	}
#else
	return NULL;
#endif
}


// ****************************************************
// * Debug strings mappings, Power IRP Minor Function *
// ****************************************************
PCHAR DbgPowerMinorFunctionString(UCHAR MinorFunction)
{
#ifdef DEBUG
	switch (MinorFunction)
	{
		case IRP_MN_SET_POWER:
			return "IRP_MN_SET_POWER";
		case IRP_MN_QUERY_POWER:
			return "IRP_MN_QUERY_POWER";
		case IRP_MN_POWER_SEQUENCE:
			return "IRP_MN_POWER_SEQUENCE";
		case IRP_MN_WAIT_WAKE:
			return "IRP_MN_WAIT_WAKE";

		default:
			return "unknown_power_irp";
	}
#else
	return NULL;
#endif
}


// ***********************************************
// * Debug strings mappings, System Power States *
// ***********************************************
PCHAR DbgSystemPowerString(SYSTEM_POWER_STATE Type)
{
#ifdef DEBUG
	switch (Type)
	{
		case PowerSystemUnspecified:
			return "PowerSystemUnspecified";
		case PowerSystemWorking:
			return "PowerSystemWorking";
		case PowerSystemSleeping1:
			return "PowerSystemSleeping1";
		case PowerSystemSleeping2:
			return "PowerSystemSleeping2";
		case PowerSystemSleeping3:
			return "PowerSystemSleeping3";
		case PowerSystemHibernate:
			return "PowerSystemHibernate";
		case PowerSystemShutdown:
			return "PowerSystemShutdown";
		case PowerSystemMaximum:
			return "PowerSystemMaximum";
		default:
			return "UnKnown System Power State";
	}
#else
	return NULL;
#endif
}


// ***********************************************
// * Debug strings mappings, Device Power States *
// ***********************************************
PCHAR DbgDevicePowerString(DEVICE_POWER_STATE Type)
{
#ifdef DEBUG
	switch (Type)
	{
		case PowerDeviceUnspecified:
			return "PowerDeviceUnspecified";
		case PowerDeviceD0:
			return "PowerDeviceD0";
		case PowerDeviceD1:
			return "PowerDeviceD1";
		case PowerDeviceD2:
			return "PowerDeviceD2";
		case PowerDeviceD3:
			return "PowerDeviceD3";
		case PowerDeviceMaximum:
			return "PowerDeviceMaximum";
		default:
			return "UnKnown Device Power State";
	}
#else
	return NULL;
#endif
}
